#ifndef PROPERTIES_H
#define PROPERTIES_H 1
#include <iostream>
#include <hash_map.h>
#include <string>
#include <sstream>
#include <fstream>
#include <vector>
#include <boost/tokenizer.hpp>

using namespace std;

template<typename keyT,typename valueT> class Properties : public hash_map<keyT,valueT, hash<keyT>, equal_to<keyT> > {
 private:
  typedef hash_map<keyT, valueT, hash<keyT>, equal_to<keyT> > MapType;
 public:
  Properties() {};
  Properties(const string& fname) {
    ifstream in(fname.c_str());
    load(in);
    in.close();
  };

  const valueT getProperty(const keyT key) const {
    typename MapType::const_iterator i=find(key);
    if(i!=end()) {
      return (i->second);
    }
  }

  bool containsKey(const keyT key) const {
    return find(key)!=end();
  }

  /*
  typename Properties<keyT,valueT>::iterator begin() {
    return map.begin();
  }

  typename Properties<keyT,valueT>::iterator end() {
    return map.end();
  }

  size_t size() const {
    return map.size();
  }

  size_t erase(const keyT& k) {
    return map.erase(k);
  }
  */
  void setProperty(const keyT key,valueT value) {
    insert(pair<const keyT,valueT>(key,value));
  }

  void load(istream& in) {
    string line;
    typedef boost::tokenizer<boost::char_separator<char> >  tokType;
    boost::char_separator<char> sep("=");
    while(getline(in,line)) {
      if(line.length()>0 && (*line.begin()!='#')) {
        tokType tabtok(line, sep);
        vector<string> v(tabtok.begin(),tabtok.end());
        if(v.size()==2) {
          // translate into appropriate types
          keyT key; valueT value;
          istringstream ks(v[0]);
          ks>>key;
          istringstream vs(v[1]);
          vs>>value;
          operator[](key)=value;
        }
      }
    }
  }

  void save(ostream& out) {
    for(typename MapType::iterator i=begin();i!=end();++i) {
      cout<<i->first<<'='<<i->second<<endl;
    }
  }

};

template<typename keyT, typename valueT>   ostream& operator<<(ostream& out, const Properties<keyT,valueT> p ) {
  p.save(out);
  return(out);
}

template<typename keyT, typename valueT>   ostream& operator<<(istream& in,Properties<keyT,valueT> p ) {
  load(in);
  return in;
}

#endif
    
