#include <stdlib.h>
#include <math.h>

#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <strstream>
#include <algorithm>
#include <string>
#include <functional>
#include <utility>

#include <stddef.h>

#include <boost/lambda/bind.hpp>
#include <boost/lambda/lambda.hpp>

#include "mcio.h"
#include "mcstats.h"
#include "mcconnectivity.h"
#include "mcdist.h"

using namespace boost::lambda;
using namespace std;


// brief help
void usage(string message, int exit_code=0) {
  if(message.length()!=0) {
    cout<<message<<"\n";
  }
  cout<<"Usage:\n mcalldist gene.set connectivity.file pvalue.expression.data ratio.expression.data\n";
  exit(exit_code);
}
void usage(int exit_code=0) {
  usage("",exit_code);
}


int main(int argc, char *argv[]) {
  if(argc<4) {
    usage();
  }
  
  string gene_set_fname(argv[1]);
  string connectivity_fname(argv[2]);
  string pvalue_expression_fname(argv[3]);
  string ratio_expression_fname(argv[4]);
  string intensity_expression_fname(argv[4]);

  string gene_distances_fname("all_gene_distances.dat");

  //  cout<<"gene set: \""<<gene_set_fname<<"\", connectivity: \""<<connectivity_fname<<"\", pvalue expression data: \""<<pvalue_expression_fname<<"\", pvalue expression data: \""<<intensity_expression_fname<<"\", intensity expression data: \""<<intensity_expression_fname<<"\"\n";
  
  vector<string> gene_names; // a list of gene names (in ID assignment order)
  //  vector< ipType > connectivity; // a list of established connectivities
  vector<string> experiment_names; // a list of experiment names

  GeneExpressionMapType pvalue_expression_map; // a map of gene name -> pvalue vector<double> for all expressed genes
  GeneExpressionMapType ratio_expression_map; // a map of gene name -> pvalue vector<double> for all expressed genes
  GeneExpressionMapType intensity_expression_map; // a map of gene name -> pvalue vector<double> for all expressed genes
  /*
  read_gene_set(gene_names,gene_set_fname);
  GeneSetType gene_set(gene_names.begin(),gene_names.end());

  //  cout<<"gene set: [";
  //  copy(gene_names.begin(),gene_names.end(),ostream_iterator<string>(cout," "));
  //  cout<<"]\n";
  cout<<"read in gene set: "<<gene_names.size()<<" genes\n"<<flush;

  cout<<"reading in connectivity data ... "<<flush;
  vector<ipType> edges;
  read_connectivity(edges,connectivity_fname);
  cout<<"done"<<endl<<flush;

  cout<<"compiling connectivity graph ... "<<flush;
  GraphType connectivity=gene_connectivity(edges,gene_names);
  cout<<"done"<<endl<<flush;
  */

  GeneSetType gene_set; // fake - we're keeping all data anyway

  // read in pvalue data
  cout<<"reading in pvalue expression data from \""<<pvalue_expression_fname<<"\" ... "<<flush;
  if(!read_expression(experiment_names,pvalue_expression_map,gene_set,pvalue_expression_fname,0,false,true)) {
    usage("ERROR: unable to read pvalue expression data from \""+pvalue_expression_fname+"\"",1);
  }
  cout<<"done"<<endl;

  // read in ratio data
  cout<<"reading in ratio expression data from \""<<ratio_expression_fname<<"\" ... "<<flush;
  if(!read_expression(experiment_names,ratio_expression_map,gene_set,ratio_expression_fname,&pvalue_expression_map,true,true)) {
    usage("ERROR: unable to read ratio expression data from \""+ratio_expression_fname+"\"",1);
  }
  cout<<"done"<<endl;


  // construct a list of expressed genes
  GeneSetType expressed_gene_set;
  for(GeneExpressionMapType::iterator i=pvalue_expression_map.begin();i!=pvalue_expression_map.end();++i) {
    expressed_gene_set.insert(i->first);
  }
  vector<string> expressed_gene_names(expressed_gene_set.begin(),expressed_gene_set.end());
  
  
  output_PValue_distance_matrix(gene_distances_fname,pvalue_expression_map,ratio_expression_map,expressed_gene_names);


  cout<<"done.\n";
}
