#include "mcconnectivity.h"

#include <boost/graph/graph_traits.hpp>
#include <boost/graph/visitors.hpp>
#include <boost/graph/johnson_all_pairs_shortest.hpp>
#include <boost/graph/connected_components.hpp>
#include <iostream>

GraphType gene_connectivity(vector<ipType>& edges,vector<string>& gene_names) {
  GraphType connectivity(edges.begin(),edges.end(),gene_names.size());
  boost::property_map<GraphType, boost::edge_weight_t>::type edgeWeight = get(boost::edge_weight, connectivity);
  for(pair<GraphType::edge_iterator,GraphType::edge_iterator> eli=boost::edges(connectivity);eli.first!=eli.second;++eli.first) {
    edgeWeight[*(eli.first)]=1;
  }
  

  //  cout<<"connectivity data: [";
  //  copy(connectivity.begin(),connectivity.end(),ostream_iterator<ipType>(cout," "));
  //  cout<<"]\n";
  //   {
  //     GraphType::edge_iterator begin,end;
  //     boost::tie(begin,end)=edges(connectivity);
  //     cout<<"read in connectivity data: "<<end-begin<<" dependencies.\n"<<flush;
  //   }

  //  for(vector<ipType>::iterator i=connectivity.begin(); i!=connectivity.end();++i) {
  //    boost::add_edge(i->first,i->second,ggraph);
  //  }
  return connectivity;
}

void calculate_gene_mdistance(vector< vector<int> >& connectivity_distance, GraphType& metabolic_connectivity,vector<string>& gene_names) {
  GraphType connectivity=metabolic_connectivity;
  vector<ipType> temp_v;  

#ifdef PATHWAY_CONNECTIVITY
  cout<<"constructing a pathway connectivity graph ... "<<flush;
  // start with the initial edge se
#ifdef CALCULATE_LINEAR_PATHWAY_PAIRS_ONLY

  GraphType pathway_connectivity(temp_v.end(),temp_v.end(),gene_names.size()); // create empty graph with n vertices
  
  //   GraphType pathway_connectivity(connectivity);
  //   // clear edges
  //   boost::graph_traits<GraphType>::vertex_iterator vi, vi_end;
  //   for (boost::tie(vi, vi_end) = vertices(pathway_connectivity); vi != vi_end; ++vi) {
  //     boost::clear_vertex(*vi,pathway_connectivity);
  //   }
  //   vector<ipType> pathway_edges;
  //   GraphType pathway_connectivity;
#else
  GraphType pathway_connectivity(connectivity);
#endif
  {
    // construct a graph consisting exclusively of the nodes
    GraphType single_connectivity(temp_v.end(),temp_v.end(),gene_names.size()); // create empty graph with n vertices
    boost::graph_traits<GraphType>::vertex_iterator vi, vi_end;
    boost::graph_traits<GraphType>::out_edge_iterator ei, ei_end;
    for (boost::tie(vi, vi_end) = vertices(connectivity); vi != vi_end; ++vi) {
      if(boost::out_degree(*vi,connectivity)<=2) {
        // get a list of edges for a given vertex
        int edge_counter=0;
        for(boost::tie(ei,ei_end)=boost::out_edges(*vi,connectivity);ei!=ei_end;++ei) {
          boost::add_edge(boost::source(*ei,connectivity),boost::target(*ei,connectivity),single_connectivity);
          edge_counter++;
        }
        if(edge_counter>2) {
          cout<<"ERROR: edge of out_degree=2 reported "<<edge_counter<<"edges!\n";
        }
      }
    }
    
    // determine connected components
    vector<int> component_assignment(num_vertices(single_connectivity));
    int num_components = boost::connected_components(single_connectivity, &component_assignment[0]);
    vector< vector<int> > components(num_components);
    for(vector<int>::iterator i=component_assignment.begin();i!=component_assignment.end();++i) {
      components[*i].push_back(i-component_assignment.begin());
    }
    
    // loop through connected components
    for(vector< vector<int> >::iterator c=components.begin();c!=components.end();++c) {
      if(c->size()>=MINIMAL_LINEAR_PATHWAY_LENGTH) { // component big enough
        // add all interconnections to the pathway connectivity
        //        cout<<"[";copy(c->begin(),c->end(),ostream_iterator<int>(cout," "));cout<<"]"<<endl<<flush;
        cout<<"[";
        for(vector<int>::const_iterator ii=c->begin();ii!=c->end();++ii) { cout<<gene_names[*ii]<<' '; }
        cout<<"]"<<endl<<flush;
#ifdef CALCULATE_LINEAR_PATHWAY_PAIRS_ONLY
        // add vertices with all their edges
        boost::graph_traits<GraphType>::out_edge_iterator ei, ei_end;
        for(vector<int>::iterator g1=c->begin();g1!=c->end();++g1) {
          // add only inner pathway edges
          if(boost::out_degree(*g1,connectivity)<=2) {
            for(boost::tie(ei,ei_end)=boost::out_edges(*g1,connectivity);ei!=ei_end;++ei) {
              //cout<<"adding edge: source "<<boost::source(*ei,connectivity)<<", target "<<boost::target(*ei,connectivity)<<endl<<flush;
              boost::add_edge(boost::source(*ei,connectivity),boost::target(*ei,connectivity),pathway_connectivity);
            }
          }
        }
        // reset edge weight
        boost::property_map<GraphType, boost::edge_weight_t>::type edgeWeight = get(boost::edge_weight, pathway_connectivity);
        for(pair<GraphType::edge_iterator,GraphType::edge_iterator> eli=boost::edges(pathway_connectivity);eli.first!=eli.second;++eli.first) {
          edgeWeight[*(eli.first)]=1;
        }
        
#else // CALCULATE_LINEAR_PATHWAY_PAIRS_ONLY
        // add all edges
        for(vector<int>::iterator g1=c->begin();g1!=c->end();++g1) {
          for(vector<int>::iterator g2=c->begin();g2!=c->end();++g2) {
            if(g1!=g2) {
              boost::add_edge(*g1,*g2,pathway_connectivity);
            }
          }
        }
#endif // CALCULATE_LINEAR_PATHWAY_PAIRS_ONLY
      }
    }
  }
  cout<<"original edges: "<<boost::num_edges(connectivity)<<" pathway edges: "<<boost::num_edges(pathway_connectivity)<<endl<<flush;
  connectivity=pathway_connectivity;
  cout<<"done."<<endl<<flush;
#endif // PATHWAY_CONNECTIVITY

  boost::johnson_all_pairs_shortest_paths(connectivity, connectivity_distance);
#ifdef USE_DIRECTED_PATHS
  // fix the matrix by considering both forward and reverse distances
  for(int i=0;i<gene_names.size();i++) {
    for(int j=0;j<gene_names.size();j++) {
      if(connectivity_distance[i][j]>connectivity_distance[j][i]) {
        connectivity_distance[i][j]=connectivity_distance[j][i];
      }
    }
  }
#endif
}
