#include "mcio.h"
#include <string>

using namespace std;

extern void usage(string,int);
extern void usage(int);

// std:: extensions
namespace std {
  template<typename charT, typename traits, typename T1, typename T2> inline std::basic_istream<charT,traits>& std::operator>> (std::basic_istream<charT,traits>& is, std::pair<T1,T2>& p) {
    T1 t1;
    T2 t2;
    is >> t1 >> t2;
    if (is)
      p = std::pair<T1,T2> (t1, t2);
    return is;
  }

  template<typename basicT>  ostream& operator<<(ostream& out,const vector<basicT> v) {
    out<<"[";
    copy(v.begin(),v.end(),ostream_iterator<basicT>(cout," "));
    out<<"]";
    return out;
  }
};


// read gene set information from file
void read_gene_set(vector<string>& gene_names, string& gene_set_fname) {
  ifstream gene_set_file(gene_set_fname.c_str());
  copy(istream_iterator<string>(gene_set_file),istream_iterator<string>(), back_inserter(gene_names));
  gene_set_file.close();
}

// read connectivity information from file
void read_connectivity(vector<ipType>& connectivity, string& connectivity_fname) {
  ifstream connectivity_file(connectivity_fname.c_str());
  copy(istream_iterator< ipType >(connectivity_file),istream_iterator< ipType >(), back_inserter(connectivity));
  connectivity_file.close();
}


// read the expression data from file
bool read_expression(vector<string>& experiment_names,GeneExpressionMapType& expression_data,GeneSetType& gene_names_set,string& expression_fname,GeneExpressionMapType* previous_expression_data,bool check,bool keep_all_data) {
  ifstream expression_file(expression_fname.c_str());
  // read the first line with the experiment names
  string line;
  if(getline(expression_file,line)) {
    typedef boost::tokenizer<boost::char_separator<char> >  tokType;
    boost::char_separator<char> sep("\t");
    tokType tabtok(line, sep);
    if(check) {
      vector<string> temp_names;
      copy(tabtok.begin(),tabtok.end(), back_inserter(temp_names));
      if(temp_names==experiment_names) {
      } else {
        cout<<"ERROR: experiment set in \""<<expression_fname<<"\"\ndiffers from previously recorded experiment names!\n";
        cout<<"previous experiment names: [";
        copy(experiment_names.begin(),experiment_names.end(),ostream_iterator<string>(cout,":::"));
        cout<<"]\n";
        cout<<expression_fname<<" experiment names: [";
        copy(temp_names.begin(),temp_names.end(),ostream_iterator<string>(cout,":::"));
        cout<<"]\n";
        return false;
      }
    } else {
      experiment_names.clear();
      copy(tabtok.begin(),tabtok.end(), back_inserter(experiment_names));
    }
  } else {
    usage("ERROR: expression data file has invalid format: unable to read the experiment list (line 1).",1);
  }

  // construct a set of registered genes
  GeneSetType existing_gene_set;
  if(check && previous_expression_data) {
    // record all genes that have expression data so far
    for(GeneExpressionMapType::iterator i=previous_expression_data->begin();i!=previous_expression_data->end();++i) {
      existing_gene_set.insert(i->first);
    }
  }
  GeneSetType new_gene_set;

  // read in the expression data
  int line_counter=1;
  while(getline(expression_file,line)) {
    line_counter++;
    istrstream ls(line.c_str());
    string geneName;
    // read off gene number first
    ls>>geneName;
    if(keep_all_data || (gene_names_set.find(geneName)!=gene_names_set.end())) {
      vector<double> v;
      // read in the experiment data
      copy(istream_iterator<double>(ls),istream_iterator<double>(),back_inserter(v));
      if(v.size()!=experiment_names.size()) {
        cout<<"ERROR: gene "<<geneName<<" in file \""<<expression_fname<<"\" has an experiment vector of size "<<v.size()<<" instead of "<<experiment_names.size()<<" (number of experiments)!\n";
        cout<<"line:\""<<line<<"\""<<endl;
        return false;
      }
      expression_data[geneName]=v;
      if(check) {
        new_gene_set.insert(geneName);
      }
    }
  }
  expression_file.close();
  if(check && previous_expression_data) {
    GeneSetType difference;
    set_symmetric_difference(existing_gene_set.begin(),existing_gene_set.end(),new_gene_set.begin(),new_gene_set.end(),inserter(difference,difference.begin()));
    if(difference.size()!=0) {
      cout<<"ERROR: file \""<<expression_fname<<"\" contained a different set of genes then the priveous file!"<<endl;
      cout<<"old gene set: [";
      copy(existing_gene_set.begin(),existing_gene_set.end(),ostream_iterator<string>(cout," "));
      cout<<"]"<<endl<<"new gene set: [";
      copy(new_gene_set.begin(),new_gene_set.end(),ostream_iterator<string>(cout," "));
      cout<<"]"<<endl<<"set difference: [";
      copy(difference.begin(),difference.end(),ostream_iterator<string>(cout," "));
      cout<<"]"<<endl<<flush;
      return false;
    }
  }
  return true;
}
