#include <stdlib.h>
#include <math.h>

#include <iostream>
#include <fstream>
#include <sstream>
#include <strstream>
#include <algorithm>
#include <string>
#include <functional>
#include <utility>

#include <stddef.h>

#include <boost/lambda/bind.hpp>
#include <boost/lambda/lambda.hpp>
#include <boost/lambda/construct.hpp>

/*
#include <gsl/gsl_multimin.h>
#include <gsl/gsl_min.h>
#include <gsl/gsl_vector.h>
*/

#include "mcio.h"
#include "mcstats.h"
#include "mcconnectivity.h"
#include "mcdist.h"
#include "npconnectivity.h"
#include "npeval.h"
#include "npdist.h"
#include "siman.h"

#define NUMBER_OF_CANDIDATES_TO_REPORT  100

using namespace boost::lambda;
using namespace std;


// brief help
void usage(string message, int exit_code=0) {
  if(message.length()!=0) {
    cout<<message<<"\n";
  }
  cout<<"Usage:\n np query.genes gene.set connectivity.file pvalue.expression.data ratio.expression.data [bianry_distances_file]\n";
  exit(exit_code);
}
void usage(int exit_code=0) {
  usage("",exit_code);
}

struct Second_more: public binary_function<const pair<string, pair<double,double> > &, const pair<string, pair<double,double> >  &, bool> {
  bool operator()(const pair<string, pair<double,double> >& p1, const pair<string, pair<double,double> >& p2) {
    return (p1.second.first > p2.second.first);
  }
};


int main(int argc, char *argv[]) {
  if(argc<6) {
    usage();
  }

  string query_gene_set_fname(argv[1]);  
  string gene_set_fname(argv[2]);
  string connectivity_fname(argv[3]);
  string pvalue_expression_fname(argv[4]);
  string ratio_expression_fname(argv[5]);
  
  string* precalculated_distances_fname=0;
  if(argc>6) {
    precalculated_distances_fname=new string(argv[6]);
  }


  string gene_distances_fname("gene_distances.dat");
  string selected_distances_fname("selected_distances.dat");

  //  cout<<"gene set: \""<<gene_set_fname<<"\", connectivity: \""<<connectivity_fname<<"\", pvalue expression data: \""<<pvalue_expression_fname<<"\", pvalue expression data: \""<<intensity_expression_fname<<"\", intensity expression data: \""<<intensity_expression_fname<<"\"\n";
  
  vector<string> gene_names; // a list of gene names (in ID assignment order)
  vector<string> query_gene_names; // a list of metabolic place names (usually U*_) for which the prediction reports should be made
  //  vector< ipType > connectivity; // a list of established connectivities
  vector<string> experiment_names; // a list of experiment names

  GeneExpressionMapType pvalue_expression_map; // a map of gene name -> pvalue vector<double> for all expressed genes
  GeneExpressionMapType ratio_expression_map; // a map of gene name -> pvalue vector<double> for all expressed genes
  GeneExpressionMapType intensity_expression_map; // a map of gene name -> pvalue vector<double> for all expressed genes

  read_gene_set(query_gene_names,query_gene_set_fname);

  read_gene_set(gene_names,gene_set_fname);
  GeneSetType gene_set(gene_names.begin(),gene_names.end());



  //  cout<<"gene set: [";
  //  copy(gene_names.begin(),gene_names.end(),ostream_iterator<string>(cout," "));
  //  cout<<"]\n";
  cout<<"read in gene set: "<<gene_names.size()<<" genes\n"<<flush;

  cout<<"reading in connectivity data ... "<<flush;
  vector<ipType> edges;
  read_connectivity(edges,connectivity_fname);
  cout<<"done"<<endl<<flush;

  cout<<"compiling connectivity graph ... "<<flush;
  GraphType connectivity=gene_connectivity(edges,gene_names);
  NPClosestConnectivity npcc(connectivity,gene_names);
  cout<<"done"<<endl<<flush;


  NPDist* npd=0;
  
  if(precalculated_distances_fname!=0) {
    npd=new NPArrayDist(*precalculated_distances_fname);
  }
  

  //  vector<double> weights(2); weights[0]=0.454092; weights[1]=1.0-weights[0]; //weights[2]=0.44;
  vector<double> weights(2); weights[0]=0.83; weights[1]=1.0-weights[0]; //weights[2]=0.4;4

  {
    cout<<"weights=[";copy(weights.begin(),weights.end(),ostream_iterator<double>(cout," ")); cout<<"]"<<endl;
    vector<double> i_ratios;
    double pie=1.0;
    for(int i=0;i<weights.size() - 1;i++) {
      double r=weights[i]/pie;
      i_ratios.push_back(r);
      pie-=weights[i];
      cout<<"ratios["<<i<<"]="<<r<<" pie="<<pie<<endl;
    }
    cout<<flush;
  }


    // read in pvalue data
    cout<<"reading in pvalue expression data from \""<<pvalue_expression_fname<<"\" ... "<<flush;
    if(!read_expression(experiment_names,pvalue_expression_map,gene_set,pvalue_expression_fname,0,false,true)) {
      usage("ERROR: unable to read pvalue expression data from \""+pvalue_expression_fname+"\"",1);
    }
    cout<<"done"<<endl;



  if(npd==0) {
    // read in ratio data
    cout<<"reading in ratio expression data from \""<<ratio_expression_fname<<"\" ... "<<flush;
    if(!read_expression(experiment_names,ratio_expression_map,gene_set,ratio_expression_fname,&pvalue_expression_map,true,true)) {
      usage("ERROR: unable to read ratio expression data from \""+ratio_expression_fname+"\"",1);
    }
    cout<<"done"<<endl;

    npd=new NPCalcDist(&pvalue_expression_map,&ratio_expression_map);
  }


  // construct a list of expressed genes
  GeneSetType expressed_gene_set;
  vector<string> expressed_gene_vector=npd->get_valid_genes();
  for(vector<string>::iterator i=expressed_gene_vector.begin();i!=expressed_gene_vector.end();++i) {
    expressed_gene_set.insert(*i);
  }

  GeneSetType nme_gene_set;
  set_difference(expressed_gene_set.begin(),expressed_gene_set.end(),gene_set.begin(),gene_set.end(),inserter(nme_gene_set,nme_gene_set.begin()));
  vector<string> nme_gene_vector(nme_gene_set.begin(),nme_gene_set.end());
  
  // do the predictions
  for(vector<string>::iterator i=query_gene_names.begin();i!=query_gene_names.end();i++) {
    cout<<"============= "<<*i<<" ============="<<endl;
    pair< pair<double,double>, vector< vector<string> > > nvar=calc_neighborhood_variability(*i,npcc,pvalue_expression_map,weights);
    cout<<"neighborhoods (variability = "<<nvar.first.first<<"):";
    // output neighborhoods
    for(vector< vector<string> >::iterator ri=nvar.second.begin();ri!=nvar.second.end();++ri) {
      if(ri->size()>0) {
        cout<<" [";
        copy(ri->begin(),ri->end(),ostream_iterator<string>(cout," "));
        cout<<']';
      }
    }
    cout<<endl;

    vector< pair<double,double> > scores=score_genes_for_position(*i,nme_gene_vector,*npd,npcc,weights);
    typedef pair<string, pair<double,double> >  NameRankPairType;
    vector<NameRankPairType> named_scores;

    transform(nme_gene_vector.begin(),nme_gene_vector.end(),scores.begin(),back_inserter(named_scores),bind(constructor< NameRankPairType >(),_1,_2));
    
    stable_sort(named_scores.begin(),named_scores.end(),Second_more());
    
    cout<<"top candidates: (gene name, distance, variance)"<<endl;
    for(int j=0;j<NUMBER_OF_CANDIDATES_TO_REPORT;j++) {
      NameRankPairType p=named_scores[j];      
      cout<<p.first<<'\t'<<p.second.first<<'\t'<<p.second.second<<endl;
    }
    
  }
  cout<<"done.\n";
}
