#include "npconnectivity.h"
#include <boost/graph/johnson_all_pairs_shortest.hpp>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/graph_traits.hpp>


NPClosestConnectivity::NPClosestConnectivity(GraphType& graph,vector<string>& gene_names): NPConnectivity(graph,gene_names), connectivity_distance(boost::num_vertices(graph),vector<int>(boost::num_vertices(graph),0)) {
  // calculate closest pair distances
  boost::johnson_all_pairs_shortest_paths(graph, connectivity_distance);
#ifdef USE_DIRECTED_PATHS
  for(int i=0;i<boost::num_vertices(graph);i++) {
    for(int j=0;j<boost::num_vertices(graph);j++) {
      if(connectivity_distance[i][j]>connectivity_distance[j][i]) {
        connectivity_distance[i][j]=connectivity_distance[j][i];
      }
    }
  }
#endif

  // record gene ids
  for(vector<string>::iterator gi=gene_names.begin();gi!=gene_names.end();++gi) {
    gene_ids[*gi]=gi-gene_names.begin();
  }
};

vector< vector<string> > NPClosestConnectivity::get_neighbours(string gene,int radius) {
  vector< vector<string> > neighbours(radius,vector<string>(0));
  GeneIdMapType::iterator id_location;
  if((id_location=gene_ids.find(gene))!=gene_ids.end()) {
    vector<int>& gene_distances=connectivity_distance[id_location->second];
    for(vector<int>::iterator g=gene_distances.begin();g!=gene_distances.end();++g) {
      int distance=*g;
      if(distance>0 && distance<=radius) {
        neighbours[distance-1].push_back(names[g-gene_distances.begin()]);
      }
    }
  } else {
    cout<<"ERROR: gene \""<<gene<<"\" is not listed in the gene_ids map."<<endl<<flush;
  }
  return neighbours;
}; 
