#ifndef NPDIST_H
#define NPDIST_H 1
#include "nptypes.h"
#include "Properties.h"
#include <string>
#include <vector>

using namespace std;

typedef hash_map<string, int, hash<string>, equal_to<string> > GeneIdMapType;

class NPDist {
 public:
  // calculate distance between two genes
  virtual pair<double,double> get_distance(const string& gene1,const string& gene2) const {}; 
  // get a list of gene names valid for distance computation
  virtual vector<string> get_valid_genes() const {}; 
};

// an instance of an NPDist class that recalculates the distance from the expression data every time
// the get_distance method is called
class NPCalcDist : public NPDist {
 protected:
  GeneExpressionMapType *pvalue_map,*ratio_map;
 public:
  NPCalcDist() {};
  NPCalcDist(GeneExpressionMapType* pvalue_expression_map,GeneExpressionMapType* ratio_exprssion_map) : pvalue_map(pvalue_expression_map), ratio_map(ratio_exprssion_map) {};
  pair<double,double> get_distance(const string& gene1,const string& gene2) const;
  vector<string> get_valid_genes() const;
};


// an instance of the NPDist class that retrieves results of get_distance from a pre-calculated array
class NPArrayDist : public NPDist {
 private:
  float *distances,*variances;
  typedef hash_map<string,int,hash<string>,equal_to<string> > GeneIdMap;
  GeneIdMap ids;
  int ngenes;
  int distance_position(int gene1, int gene2) const;
 protected:
  vector<string> gene_names;
 public:
  NPArrayDist(string& precalculated_distances_file_name);
  pair<double,double> get_distance(const string& gene1,const string& gene2) const;
  vector<string> get_valid_genes() const;
};

class NPConditionDist : public NPDist {
 private:
  Properties<string,int> gdel_mapping;
  GeneIdMapType gidm;
 protected:
  GeneExpressionMapType pvalue_map;
  GeneExpressionMapType ratio_map;
 public:
  NPConditionDist(string& pvalue_fname,string& ratio_fname,string& gdel_mapping_fname);
  pair<double,double> get_distance(const string& gene1,const string& gene2) const;
  vector<string> get_valid_genes() const;
  double get_pvalue(const string& condition_name,const string& gene_name) const;
};

class NPShuffledArrayDist : public NPArrayDist {
 private:
  typedef hash_map<string, string,hash<string>,equal_to<string> > GeneNameMapType;
  GeneNameMapType shuffled_gene_name;
 public:
  NPShuffledArrayDist(string& precalculated_distances_file_name);
  pair<double,double> get_distance(const string& gene1,const string& gene2) const;
  vector<string> get_valid_genes() const;
  void shuffle_genes();
};

void calculate_PValue_distance_matrix(vector< vector< pair<double,double> > >& distances,NPDist& npd,vector<string>& gene_names);

#endif
