#ifndef NPEVAL_H
#define NPEVAL_H 1
#include <gsl/gsl_vector.h>

#include "nptypes.h"
#include "npconnectivity.h"
#include "npdist.h"


#define NEIGHBOURHOOD_RADIUS 4
#define LARGE_COST 1e20

#define POWER_FACTOR 12.0

typedef struct {
  vector<string> *ptr_metabolic_genes;
  vector<string> *ptr_nme_genes;
  NPDist  *ptr_npd;
  NPConnectivity  *ptr_npc;
} self_rank_fitness_controls;

// comparator by the first number in the pair
struct First_less: public binary_function<const pair<int,string> &, const pair<int,string> &, bool> {
  bool operator()(const pair<int,string>& p1, const pair<int,string>& p2) {
    return (p1.first < p2.first);
  }
};


pair<double,double> placement_cost_function(const vector< vector< pair<double,double> > >& neighbour_distances,const vector<double>& weights);
vector< pair<double,double> > compute_distance_to_gene(const string& gene,const vector<string>& genes,NPDist& npd);
vector< vector< pair<double,double> > > compute_distances_to_gene(const string& gene,const vector< vector<string> >& genes,NPDist& npd);
vector< pair<double,double> > score_genes_for_position(const string& positionGene, const vector<string>& genes,NPDist& npd,NPConnectivity& npc,const vector<double>& weights);
pair<int,pair<double,double> > rank_gene(const string& metabolic_gene,const vector<string>& gene_vector,NPDist& npd,NPConnectivity& npc,const vector<double>& weights);
double self_rank_fitness(const gsl_vector *v, void *params);
double self_logrank_fitness(const gsl_vector *v, void *params);
double self_max50_fitness(const gsl_vector *v, void *params);
double self_psum_fitness(const gsl_vector *v, void *params);
pair< pair<double,double>, vector< vector<string> > > calc_neighborhood_variability(const string& positionGene,NPConnectivity& npc,const GeneExpressionMapType& pvalue_expression_map,const vector<double>& weights);
// pair< vector< vector<double> >, vector< vector<string> > > calc_neighborhood_variability_vector(const string& positionGene,NPConnectivity& npc,const GeneExpressionMapType& pvalue_expression_map,const int radius);
#endif
