#include <stdlib.h>
#include <math.h>

#include <iostream>
#include <fstream>
#include <sstream>
#include <strstream>
#include <algorithm>
#include <string>
#include <functional>
#include <utility>

#include <stddef.h>

#include <boost/lambda/bind.hpp>
#include <boost/lambda/lambda.hpp>


#include <gsl/gsl_multimin.h>
#include <gsl/gsl_min.h>
#include <gsl/gsl_vector.h>


#include "mcio.h"
#include "mcstats.h"
#include "mcconnectivity.h"
#include "mcdist.h"
#include "npconnectivity.h"
#include "npeval.h"
#include "npdist.h"
#include "siman.h"

using namespace boost::lambda;
using namespace std;


// brief help
void usage(string message, int exit_code=0) {
  if(message.length()!=0) {
    cout<<message<<"\n";
  }
  cout<<"Usage:\n npfit gene.set connectivity.file pvalue.expression.data [bianry_distances_file]\n";
  exit(exit_code);
}
void usage(int exit_code=0) {
  usage("",exit_code);
}


int main(int argc, char *argv[]) {
  if(argc<5) {
    usage();
  }
  
  string gene_set_fname(argv[1]);
  string connectivity_fname(argv[2]);
  string pvalue_expression_fname(argv[3]);
  //  string ratio_expression_fname(argv[4]);
  //  string intensity_expression_fname(argv[5]);
  string precalculated_distances_fname=argv[4];  

  string gene_distances_fname("gene_distances.dat");
  string selected_distances_fname("selected_distances.dat");

  //  cout<<"gene set: \""<<gene_set_fname<<"\", connectivity: \""<<connectivity_fname<<"\", pvalue expression data: \""<<pvalue_expression_fname<<"\", pvalue expression data: \""<<intensity_expression_fname<<"\", intensity expression data: \""<<intensity_expression_fname<<"\"\n";
  
  vector<string> gene_names; // a list of gene names (in ID assignment order)
  //  vector< ipType > connectivity; // a list of established connectivities
  vector<string> experiment_names; // a list of experiment names

  GeneExpressionMapType pvalue_expression_map; // a map of gene name -> pvalue vector<double> for all expressed genes

  read_gene_set(gene_names,gene_set_fname);
  GeneSetType gene_set(gene_names.begin(),gene_names.end());



  //  cout<<"gene set: [";
  //  copy(gene_names.begin(),gene_names.end(),ostream_iterator<string>(cout," "));
  //  cout<<"]\n";
  cout<<"read in gene set: "<<gene_names.size()<<" genes\n"<<flush;

  cout<<"reading in connectivity data ... "<<flush;
  vector<ipType> edges;
  read_connectivity(edges,connectivity_fname);
  cout<<"done"<<endl<<flush;
  cout<<"compiling connectivity graph ... "<<flush;
  GraphType connectivity=gene_connectivity(edges,gene_names);
  NPClosestConnectivity npcc(connectivity,gene_names);
  cout<<"done"<<endl<<flush;


  NPDist* npd=new NPArrayDist(precalculated_distances_fname);

  // vector<double> weights(3); weights[0]=0.7; weights[1]=0.7*(1.0-weights[0]); weights[2]=1.0-weights[0]-weights[1];
  // vector<double> weights(2); weights[0]=0.78125; weights[1]=1.0-weights[0]; regrank p12
  // vector<double> weights(2); weights[0]=0.83125; weights[1]=1.0-weights[0]; logrank p12
  // vector<double> weights(3); weights[0]=0.892654; weights[1]=0.9882875*(1.0-weights[0]); weights[2]=1.0-weights[0]-weights[1];

  //vector<double> weights(2); weights[0]=0.9; weights[1]=1.0-weights[0]; 
  vector<double> weights(3); weights[0]=0.8; weights[1]=0.8*(1.0-weights[0]); weights[2]=1.0-weights[0]-weights[1];
  // vector<double> weights(3); weights[0]=0.7; weights[1]=0.7*(1.0-weights[0]); weights[2]=1.0-weights[0]-weights[1];
 
  // vector<double> weights(3); weights[0]=0.77492; weights[1]=0.999414*(1.0-weights[0]); weights[2]=1.0-weights[0]-weights[1];
  //  vector<double> weights(2); weights[0]=0.454092; weights[1]=1.0-weights[0]; //weights[2]=0.44;
  // optimal for ratio + pow 12
  // vector<double> weights(3); weights[0]=0.841; weights[1]=1.0-weights[0]; // weights[2]=0.44;

  // zscore pow12
  // vector<double> weights(3); weights[0]=0.74776; weights[1]=0.88621*(1.0-weights[0]); weights[2]=1.0-weights[0]-weights[1];


  //  vector<double> weights(4); weights[0]=0.74776; weights[1]=0.88621*(1.0-weights[0]); weights[2]=0.88*(1.0-weights[0]-weights[1]); weights[3]=1.0-weights[0]-weights[1]-weights[2];

  // zscore pow12 directed:  0.80584 0.875959
  // vector<double> weights(3); weights[0]=0.70584; weights[1]=0.875959*(1.0-weights[0]); weights[2]=1.0-weights[0]-weights[1];


  // vector<double> weights(2); weights[0]=0.841; weights[1]=0.8;
  // vector<double> weights(3); weights[0]=0.5; weights[1]=0.3; weights[2]=1.0-weights[0]-weights[1];
  // vector<double> weights(3); weights[0]=0.841; weights[1]=0.05; weights[2]=1.0-weights[0]-weights[1];

  // optimal for zscores pow2
  // vector<double> weights(2); weights[0]=0.716; weights[1]=1.0-weights[0]; //weights[2]=0.4;4
  // vector<double> weights(3); weights[0]=0.716; weights[1]=0.1; weights[2]=1.0-weights[0]-weights[1];


  {
    cout<<"weights=[";copy(weights.begin(),weights.end(),ostream_iterator<double>(cout," ")); cout<<"]"<<endl;
    vector<double> i_ratios;
    double pie=1.0;
    for(int i=0;i<weights.size() - 1;i++) {
      double r=weights[i]/pie;
      i_ratios.push_back(r);
      pie-=weights[i];
      cout<<"ratios["<<i<<"]="<<r<<" pie="<<pie<<endl;
    }
    cout<<flush;
  }



  // read in pvalue data
  cout<<"reading in pvalue expression data from \""<<pvalue_expression_fname<<"\" ... "<<flush;
  if(!read_expression(experiment_names,pvalue_expression_map,gene_set,pvalue_expression_fname,0,false,true)) {
    usage("ERROR: unable to read pvalue expression data from \""+pvalue_expression_fname+"\"",1);
  }
  cout<<"done"<<endl;

  /*  
  {
    // testing gene/experiment combinations for something else
    vector< pair<string,int> > t;
    t.push_back(pair<string,int>(string("YAR015W"),1));
    t.push_back(pair<string,int>(string("YLR028C"),2));
    t.push_back(pair<string,int>(string("YOR128C"),3));
    t.push_back(pair<string,int>(string("YER073W"),6));
    t.push_back(pair<string,int>(string("YER069W"),12));
    t.push_back(pair<string,int>(string("YER061C"),25));
    t.push_back(pair<string,int>(string("YOR236W"),39));
    t.push_back(pair<string,int>(string("YHR208W"),45));
    t.push_back(pair<string,int>(string("YBR176W"),50));
    t.push_back(pair<string,int>(string("YMR202W"),54));
    t.push_back(pair<string,int>(string("YLR056W"),55));
    t.push_back(pair<string,int>(string("YKL004W"),278));
    t.push_back(pair<string,int>(string("YGL008C"),285));

    
    for(vector< pair<string,int> >::iterator i=t.begin();i!=t.end();++i) {
      GeneExpressionMapType::const_iterator pv=pvalue_expression_map.find(i->first);
      cout<<"gene "<<i->first;
      if(pv!=pvalue_expression_map.end()) {
        cout<<" experiment "<<i->second<<" pvalue"<<(pv->second)[i->second-1]<<endl;
      } else {
        cout<<" is not expressed !"<<endl;
      }
    }
  }
  */

//   // read in ratio data
//   cout<<"reading in ratio expression data from \""<<ratio_expression_fname<<"\" ... "<<flush;
//   if(!read_expression(experiment_names,ratio_expression_map,gene_set,ratio_expression_fname,&pvalue_expression_map,true,true)) {
//     usage("ERROR: unable to read ratio expression data from \""+ratio_expression_fname+"\"",1);
//   }
//   cout<<"done"<<endl;

//   NPCalcDist npcd(&pvalue_expression_map,&ratio_expression_map);
//   if(npd==0) { npd=&npcd; }


  // construct a list of expressed genes
  GeneSetType expressed_gene_set;
  for(GeneExpressionMapType::iterator i=pvalue_expression_map.begin();i!=pvalue_expression_map.end();++i) {
    expressed_gene_set.insert(i->first);
  }

  /*
  {
    GeneSetType difference;
    set_difference(gene_set.begin(),gene_set.end(),expressed_gene_set.begin(),expressed_gene_set.end(),inserter(difference,difference.begin()));

    if(difference.size()!=0) {
      cout<<"genes that were not expressed: [";
      copy(difference.begin(),difference.end(),ostream_iterator<string>(cout," "));
      cout<<"]"<<endl<<flush;
    } else {
      cout<<"all of the above genes were expressed."<<endl<<flush;
    }
  }
  */

  // non-metabolic expressed gene set (candidate genes)
  GeneSetType nme_gene_set;
  set_difference(expressed_gene_set.begin(),expressed_gene_set.end(),gene_set.begin(),gene_set.end(),inserter(nme_gene_set,nme_gene_set.begin()));

  /*
  if(nme_gene_set.size()!=0) {
    cout<<"non-metabolic expressed genes: [";
    copy(nme_gene_set.begin(),nme_gene_set.end(),ostream_iterator<string>(cout," "));
    cout<<"]"<<endl<<flush;
  } else {
    cout<<"all expressed genes were metabolic."<<endl<<flush;
  }
  */

  // expressed metabolic gene set (test genes)
  GeneSetType expressed_mgene_set;
  set_intersection(expressed_gene_set.begin(),expressed_gene_set.end(),gene_set.begin(),gene_set.end(),inserter(expressed_mgene_set,expressed_mgene_set.begin()));

  /*  
  if(nme_gene_set.size()!=0) {
    cout<<"metabolic expressed genes: [";
    copy(expressed_mgene_set.begin(),expressed_mgene_set.end(),ostream_iterator<string>(cout," "));
    cout<<"]"<<endl<<flush;
  } else {
    cout<<"there are no metabolic genes present with expression data.."<<endl<<flush;
  }
  */

  { // calculate neigborhood variability for all metabolic places
    ofstream vout("neighborhood.variability");
    vout<<"place\tnvar\tsvar\tneighborhood genes\n";
    for(GeneSetType::iterator pi=gene_set.begin();pi!=gene_set.end();++pi) {
      pair< pair<double,double>, vector< vector<string> > > nvar=calc_neighborhood_variability(*pi,npcc,pvalue_expression_map,weights);
      vout<<*pi<<'\t'<<nvar.first.first<<'\t'<<nvar.first.second<<'\t';
      // output neighborhoods
      for(vector< vector<string> >::iterator ri=nvar.second.begin();ri!=nvar.second.end();++ri) {
        if(ri->size()>0) {
          vout<<" (";
          copy(ri->begin(),ri->end(),ostream_iterator<string>(vout," "));
          vout<<')';
        }
      }
      vout<<endl;
    }
    vout.close();
  }


  {


    int SR_N_RANDOM=expressed_mgene_set.size(); // size of the random gene sample
    int SR_N_TOP_RANKING=300; // number of the top ranking genes to optimize on

    // take a random subset of genes
    srand48(time(0));
    vector<string> mgenes_shuffled;
    copy(expressed_mgene_set.begin(),expressed_mgene_set.end(),back_inserter(mgenes_shuffled));
    random_shuffle(mgenes_shuffled.begin(),mgenes_shuffled.end());
    vector<string> mgenes_subset;
    copy(mgenes_shuffled.begin(),mgenes_shuffled.begin()+SR_N_RANDOM,back_inserter(mgenes_subset));
    

    // construct a gene vector
    vector<string> nme_gene_vector(nme_gene_set.begin(),nme_gene_set.end());
    cout<<"number of non-metabolic expressed ORFs = "<<nme_gene_vector.size()<<endl<<flush;
    
    ofstream fout("selfrank.run");
    fout<<"gene\trank\tdistance\tvar\tnvar\tsvar\n";
    vector< pair<int,string> > rank_run;

    vector<int> top_counts(4,0);
    for(vector<string>::iterator emgi=mgenes_subset.begin();emgi!=mgenes_subset.end();++emgi) {
      //      cout<<*emgi<<" ranks npad:"<<rank_gene(*emgi,nme_gene_set,*npd,npcc,weights)<<flush<<" npcd:"<<rank_gene(*emgi,nme_gene_set,npcd,npcc,weights)<<endl<<flush;
      pair<int,pair<double,double> > r=rank_gene(*emgi,nme_gene_vector,*npd,npcc,weights);

      // determine neighborhood variability
      pair<double,double> nvar=calc_neighborhood_variability(*emgi,npcc,pvalue_expression_map,weights).first;

      fout<<*emgi<<'\t'<<r.first<<'\t'<<r.second.first<<'\t'<<r.second.second<<'\t'<<nvar.first<<'\t'<<nvar.second<<endl<<flush;
      // cout<<*emgi<<" ranks "<<r<<endl<<flush;

      if(r.first > 0) {
        if(r.first < 10) {
          top_counts[0]++;
        }
        if(r.first < 50) {
          top_counts[1]++;
        }
        if(r.first < 100) {
          top_counts[2]++;
        }
        if(r.first < 200) {
          top_counts[3]++;
        }
      }

      if(r.first>0) {
        rank_run.push_back(pair<int,string>(r.first,*emgi));
      }
    }
    fout.close();
    cout<<"top count=[";copy(top_counts.begin(),top_counts.end(),ostream_iterator<int>(cout," ")); cout<<"]"<<endl;


    /* 
    { // incremental step reports
      for(double ratio=atof(argv[7]);ratio+=atof(argv[8]);ratio<atof(argv[9])) {
        vector<int> top_counts_inc(4,0);
        int valid_rank_count=0;
        double mean_rank=0.0;
        
        vector<double> local_weights(2);
        local_weights[0]=ratio;
        local_weights[1]=1.0-ratio;

        for(vector<string>::iterator emgi=mgenes_subset.begin();emgi!=mgenes_subset.end();++emgi) {
          //      cout<<*emgi<<" ranks npad:"<<rank_gene(*emgi,nme_gene_set,*npd,npcc,weights)<<flush<<" npcd:"<<rank_gene(*emgi,nme_gene_set,npcd,npcc,weights)<<endl<<flush;
          pair<int,pair<double,double> > r=rank_gene(*emgi,nme_gene_vector,*npd,npcc,local_weights);
          
          if(r.first > 0) {
            valid_rank_count++;
            mean_rank+=(double)r.first;

            if(r.first < 10) {
              top_counts_inc[0]++;
            }
            if(r.first < 50) {
              top_counts_inc[1]++;
            }
            if(r.first < 100) {
              top_counts_inc[2]++;
            }
            if(r.first < 200) {
              top_counts_inc[3]++;
            }
          }
        }
        fout.close();
        cout<<"weights=[";copy(local_weights.begin(),local_weights.end(),ostream_iterator<double>(cout," ")); cout<<"]";
        cout<<" [";copy(top_counts_inc.begin(),top_counts_inc.end(),ostream_iterator<int>(cout," ")); cout<<"]";
        cout<<" m="<<mean_rank/((double)valid_rank_count)<<endl;
      }
    }



    exit(0);
    */
    
    // use top N genes from the subset for the optimization
    /*
    stable_sort(rank_run.begin(),rank_run.end(),First_less());
    
    vector<string> top_ranking_metabolic_genes;
    cout<<"top ranking genes: [";
    for(int i=0;i<SR_N_TOP_RANKING;i++) {
      cout<<rank_run[i]<<" ";
      top_ranking_metabolic_genes.push_back(rank_run[i].second);
    }
    cout<<"]"<<endl<<flush;
    */

    vector<string> top_ranking_metabolic_genes=mgenes_subset; // use the entire subset for the optimization

    /*
    { // simulated annealing
      // set up params

      // how many points do we try before stepping 
#define N_TRIES 30
      
      // how many iterations for each T? 
#define ITERS_FIXED_T 3
      
      // max step size in random walk
#define STEP_SIZE 0.4
      
      // Boltzmann constant 
#define K 1.0             
      
      // initial temperature 
#define T_INITIAL 0.002         
      
      // damping factor for temperature 
#define MU_T 1.005              
#define T_MIN 2.0e-6
      
      gsl_siman_params_t params 
        = {N_TRIES, ITERS_FIXED_T, STEP_SIZE,
           K, T_INITIAL, MU_T, T_MIN};
      
      
      // initial configuration 
      Siman_control_struct_type i_c;

      i_c.ptr_metabolic_genes=&top_ranking_metabolic_genes;
      i_c.ptr_nme_genes=&nme_gene_vector;
      i_c.ptr_npd=npd;
      i_c.ptr_npc=&npcc;


      vector<double> i_ratios;
      double pie=1.0;
      cout<<"weights=[";copy(weights.begin(),weights.end(),ostream_iterator<double>(cout," ")); cout<<"]"<<endl;
      for(int i=0;i<weights.size() - 1;i++) {
        double r=weights[i]/pie;
        i_ratios.push_back(r);
        pie-=weights[i];
        cout<<"ratios["<<i<<"]="<<r<<" pie="<<pie<<endl;
      }

      i_c.ratios=&i_ratios;
      
      // set up random number genrator
      gsl_rng_env_setup();
      gsl_rng *r=gsl_rng_alloc(gsl_rng_default);

      // run simulated annealing 
      gsl_siman_solve(r, &i_c, siman_f, siman_step, siman_metric, siman_print,
                      siman_copy, siman_copy_construct, siman_destroy, 
                      0, params);

    } // simulated annealing

    */    


    { // gradient minimization
      double SR_STEP_SIZE=0.1;      
      size_t np = weights.size()-1;

      self_rank_fitness_controls par;
      par.ptr_metabolic_genes=&top_ranking_metabolic_genes;
      par.ptr_nme_genes=&nme_gene_vector;
      par.ptr_npd=npd;
      par.ptr_npc=&npcc;
    
      const gsl_multimin_fminimizer_type *T = gsl_multimin_fminimizer_nmsimplex;
      gsl_multimin_fminimizer *s = NULL;
      gsl_vector *ss, *x;
      gsl_multimin_function minex_func;
    
      size_t iter = 0, i;
      int rval = GSL_CONTINUE;
      int status = GSL_SUCCESS;
      double ssval;

      // initial vertex size vector
    
      ss = gsl_vector_alloc (np);
    
      if (ss == NULL) {
        GSL_ERROR_VAL ("failed to allocate space for ss", GSL_ENOMEM, 0);
      }
  
      gsl_vector_set_all (ss, SR_STEP_SIZE);

      // starting point 

      x = gsl_vector_alloc (np);
    
      if (x == NULL) {
        gsl_vector_free(ss);
        GSL_ERROR_VAL ("failed to allocate space for x", GSL_ENOMEM, 0);
      }

      double pie=1.0;
      for(int i=0;i<np;i++) {
        double r=weights[i]/pie;
        gsl_vector_set (x, i, r);
        pie-=weights[i];
        cout<<"r["<<i<<"]="<<r<<endl;
      }

      // initialize method and iterate
    
      minex_func.f = &self_logrank_fitness;
      minex_func.n = np;
      minex_func.params = (void *)&par;
    
      s = gsl_multimin_fminimizer_alloc (T, np);
      gsl_multimin_fminimizer_set (s, &minex_func, x, ss);
    
      while (rval == GSL_CONTINUE)  {
        iter++;
        status = gsl_multimin_fminimizer_iterate(s);
      
        if (status) 
          break;
      
        rval = gsl_multimin_test_size (gsl_multimin_fminimizer_size (s),1e-3);
        ssval = gsl_multimin_fminimizer_size (s);

        if (rval == GSL_SUCCESS)
          cout<<"converged to minimum at"<<endl;
      
        cout<<iter<<' ';
        for (i = 0; i < np; i++) {
          cout<<gsl_vector_get (s->x, i)<<" ";
        }
        cout<<"\tf()="<<s->fval<<" ssize="<<ssval<<endl<<flush;
      }

      // set the weights vector accordingly

      pie=1.0;
      weights.clear();
      for(i=0;i<s->x->size;i++) {
        double ratio=gsl_vector_get(s->x,i);
        // determine the slice
        double slice=pie*ratio;
        weights.push_back(slice);
        pie=pie-slice;
      }
      weights.push_back(pie);
      cout<<"weights=[";copy(weights.begin(),weights.end(),ostream_iterator<double>(cout," ")); cout<<"]"<<endl;
      
      gsl_vector_free(x);
      gsl_vector_free(ss);
      gsl_multimin_fminimizer_free (s);

    } // gradient minimization 


    {
      // do another selfrank run

      ofstream fout("selfrank2.run");
      fout<<"gene\trank\tdistance\tvar\tnvar\tsvar\n";
      vector< pair<int,string> > rank_run;


      vector<int> top_counts2(4,0);
      vector<string> emgenes_vector(expressed_mgene_set.begin(),expressed_mgene_set.end());
      for(vector<string>::iterator emgi=emgenes_vector.begin();emgi!=emgenes_vector.end();++emgi) {
        //      cout<<*emgi<<" ranks npad:"<<rank_gene(*emgi,nme_gene_set,*npd,npcc,weights)<<flush<<" npcd:"<<rank_gene(*emgi,nme_gene_set,npcd,npcc,weights)<<endl<<flush;
        pair<int,pair<double,double> > r=rank_gene(*emgi,nme_gene_vector,*npd,npcc,weights);
        
        // determine neighborhood variability
        pair<double,double> nvar=calc_neighborhood_variability(*emgi,npcc,pvalue_expression_map,weights).first;
        
        fout<<*emgi<<'\t'<<r.first<<'\t'<<r.second.first<<'\t'<<r.second.second<<'\t'<<nvar.first<<'\t'<<nvar.second<<endl<<flush;
        // cout<<*emgi<<" ranks "<<r<<endl<<flush;

        if(r.first>0) {
          if(r.first < 10) {
            top_counts2[0]++;
          }
          if(r.first < 50) {
            top_counts2[1]++;
          }
          if(r.first < 100) {
            top_counts2[2]++;
          }
          if(r.first < 200) {
            top_counts2[3]++;
          }
        }
        
        if(r.first>0) {
          rank_run.push_back(pair<int,string>(r.first,*emgi));
        }
      }
      fout.close();
      cout<<"top count=";copy(top_counts2.begin(),top_counts2.end(),ostream_iterator<int>(cout," ")); cout<<"]"<<endl;      
    }

  }

  cout<<"done.\n";
}
