#include "siman.h"
#include <math.h>
#include <iostream>
#include <iterator>
#include "npeval.h"

// function being minimized
double siman_f(void *xp) {
  Siman_control_struct_type *controls=(Siman_control_struct_type*)xp;
  // reconstruct the weights vector
  vector<double> weights;
  double pie=1.0;
  for(vector<double>::iterator ri=controls->ratios->begin();ri!=controls->ratios->end();++ri) {
    if((*ri) >1.0 || (*ri) < 0.0) {
      return LARGE_COST;
    }
    double slice=pie*(*ri);
    
    weights.push_back(slice);
    pie=pie-slice;
  }
  weights.push_back(pie); // last weight
  
  if(pie<0.0) { 
    cout<<"suman_f(): ERROR: negative amount of pie left for a valid ratios=[";
    copy(controls->ratios->begin(),controls->ratios->end(),ostream_iterator<double>(cout," "));
    cout<<"] and weights=[";
    copy(weights.begin(),weights.end(),ostream_iterator<double>(cout," "));
    cout<<"]"<<endl<<flush;
  }

  // self-rank all given metabolic genes
  int rank_sum=0;
  for(vector<string>::iterator mgi=controls->ptr_metabolic_genes->begin();mgi!=controls->ptr_metabolic_genes->end();++mgi) {
    rank_sum+=rank_gene(*mgi,*(controls->ptr_nme_genes),*(controls->ptr_npd),*(controls->ptr_npc),weights).first;
  }
  double mean_rank= ((double)rank_sum)/((double)controls->ptr_metabolic_genes->size());
  cout<<"siman_f("; copy(controls->ratios->begin(),controls->ratios->end(),ostream_iterator<double>(cout,",")); cout<<")="<<mean_rank<<endl<<flush;
  return mean_rank;
}
      
// distance function
double siman_metric(void *xp, void *yp){
  Siman_control_struct_type *cx=(Siman_control_struct_type*)xp;

  Siman_control_struct_type *cy=(Siman_control_struct_type*)xp;

  double dist=0.0;
  for(int i=0;i<cx->ratios->size();i++) {
    dist+=fabs(cx->ratios->operator[](i) - cy->ratios->operator[](i));
  }
  
  return dist;
}

// step construction function 
void siman_step(const gsl_rng * r, void *xp, double step_size) {
  Siman_control_struct_type *cx=(Siman_control_struct_type*)xp;

  // absolute step distance
  double distance=2.0*step_size*gsl_rng_uniform(r)-step_size;
  // direction
  vector<double> direction;
  double d_sq_size=0.0;
  for(int i=cx->ratios->size();i>0;i--) {
    double rd=gsl_rng_uniform(r);
    direction.push_back(rd);
    d_sq_size+=rd*rd;
  }
  
  // scale direction
  double scaling=distance/sqrt(d_sq_size);

  for(int i=cx->ratios->size()-1;i>=0;i--) {
    (cx->ratios->operator[](i))+=scaling*direction[i];
  }  
}

// configuration reporting function 
void siman_print(void *xp)  {
  Siman_control_struct_type *c=(Siman_control_struct_type*)xp;
  cout<<" :";
  copy(c->ratios->begin(),c->ratios->end(),ostream_iterator<double>(cout," "));
  cout<<":"<<flush;
}

// control structure copy constructor
void* siman_copy_construct(void *source) {
  Siman_control_struct_type *old_c=(Siman_control_struct_type*)source;  
  Siman_control_struct_type *new_c=(Siman_control_struct_type*)malloc(sizeof(Siman_control_struct_type));
  memcpy(new_c,old_c,sizeof(Siman_control_struct_type));
  // redefine the ratio vector
  new_c->ratios=new vector<double>(*(old_c->ratios));
  return new_c;
}


// control structure copy
void siman_copy(void *source,void *dest) {
  Siman_control_struct_type *old_c=(Siman_control_struct_type*)source;  
  Siman_control_struct_type *new_c=(Siman_control_struct_type*)dest;

  // clear out dest ratios vector
  delete(new_c->ratios);
  memcpy(new_c,old_c,sizeof(Siman_control_struct_type));
  // redefine the ratio vector
  new_c->ratios=new vector<double>(*(old_c->ratios));
}

// destructor
void siman_destroy(void *source) {
  Siman_control_struct_type *old_c=(Siman_control_struct_type*)source;
  delete(old_c->ratios); old_c->ratios=0;
  free(old_c);
}
  

